open Ast

(* AST-based optimizations *)

type opt_context = { u : int;
		     b : bool;
		     em : bool;
		     s : bool;
		     i : bool;
		     tt : bool;
		     col : color;
		     sz : int }

let root_ctx = { u = 0; 
		 b = false;
		 em = false;
		 s = false;
		 i = false;
		 tt = false;
		 col = Black;
		 sz = 0;
	       }

let is_plain ctx = (ctx.u = 0) && (not (ctx.b || ctx.em || ctx.i || ctx.tt))

(* a node is an irrelevant_em when there are no emphasized printing characters
   inside it *)


let is_space = function
	  ' ' -> true
	| '\t' -> true
	| '\n' -> true
	| '\r' -> true
	| _ -> false

let is_all_space s = 
  let len = (String.length s) in
  let rec check n = 
    if n = len then true else
      if (is_space (String.get s n))
      then check (n+1)
      else false
  in
    check 0

let rec irrelevant_em emph_status x = 
  match x with 
      Txt str -> (not emph_status) || (is_all_space str)
    | Seq sz -> List.for_all (fun x' -> irrelevant_em emph_status x') sz 
    | Prop (EM, sub) -> irrelevant_em (not emph_status) sub
    | Prop (S, sub) -> irrelevant_em false sub
    | Empty -> true
    | _ -> false

let rec trash_whitespace = function
    Txt str -> is_all_space str
  | Prop (U, _) -> false
  | Prop (_,x) -> trash_whitespace x
  | Seq sz -> List.for_all trash_whitespace sz
  | Empty -> true

let rec begins_space = function
    Txt str -> (String.length str = 0) || is_space (String.get str 0)
  | Prop (U, _) -> false
  | Prop (_,x) -> begins_space x
  | Seq (s::sz) -> begins_space s
  | Seq [] -> true
  | Empty -> true

let rec ends_space = function
    Txt str -> (String.length str = 0) || is_space (String.get str (String.length str - 1))
  | Prop (U, _) -> false
  | Prop (_,x) -> ends_space x
  | Seq [] -> true
  | Seq sz -> ends_space (List.hd (List.rev sz))
  | Empty -> true

let squeeze_spaces s =
  let p = ref 0 in
  let n = String.create (String.length s) in
  for i = 0 to (String.length s) - 1 do
    n.[!p] <- s.[i];
    begin
      if (i > 0) && (is_space s.[i-1] && is_space s.[i]) 
      then ()
      else p := (!p + 1)
    end
  done;
    String.sub n 0 !p

(* we can merge non-overlapping, non-adjoining property extents into overlapping
   property extents _only_ if the properties are mutually exclusive. *)

let props_excl = function
    (Col x, Col y) -> true
  | (Sz x, Sz y) -> true
  | (_,_) -> false

(* this is for merging (near-) neighbours *)

let rec merge = function
    ((Prop (p,a)) :: (Prop(q,b)) :: rest) 
    when p=q -> (Prop (p, (Seq [a; b]))) :: (merge rest)      

  | ((Prop (p,a)) :: ((Prop(q,_)) as b) :: (Prop(r,c)) :: rest) 
      when (p=r && props_excl (p,q)) -> merge ((Prop (p, (Seq [a; b; c]))) :: rest)

  | s1 :: s2 :: sz 
      when (trash_whitespace s1) && (begins_space s2)
	-> merge (s2 :: sz)

  | s1 :: s2 :: sz 
      when (ends_space s1) && (trash_whitespace s2)
	-> merge (s1 :: sz)
      
  | x::xs -> x :: (merge xs)
  | [] -> []



let rec opt ctx d = match d with
    
    (* plainness-identity law *)
    
    Prop (x,(Prop (PL, y))) 
      when List.exists (fun k -> k = x) [B; EM; I; PL; S; TT; U]  
      -> Prop (PL, (opt ctx y))
	
  (* common section & whitespace merging *)

  | Seq sz -> Seq (merge (List.map (opt ctx) sz))
      
  (* idempotence rules *)

  | Prop(U,x) when ctx.u > 2 -> opt ctx x
  | Prop(U,x) -> Prop(U,opt {ctx with u=(ctx.u+1)} x)
  | Prop (B,x) when ctx.b -> opt ctx x
  | Prop (B,x) -> Prop(B,opt {ctx with b=true} x)
  | Prop (I,x) when ctx.i -> opt ctx x
  | Prop (I,x)  -> Prop(I,opt {ctx with i=true} x)
  | Prop (TT,x) when ctx.tt -> opt ctx x
  | Prop (TT,x) -> Prop(TT,opt {ctx with tt=true} x)
  | Prop (PL,x) when (is_plain ctx) -> opt ctx x
  | Prop (PL,x) -> Prop(PL,opt {root_ctx with 
				  col=ctx.col; sz=ctx.sz} x)
  | Prop (Col c,x) when ctx.col = c -> opt ctx x
  | Prop (Col c,x) -> Prop(Col c,opt {ctx with col=c} x)
  | Prop (Sz n,x) when ctx.sz = n -> opt ctx x
  | Prop (Sz n,x) -> Prop(Sz n,opt {ctx with sz=n} x)

  (* EM / S elimination *)
      
  | Prop (S,x) when ctx.s -> opt ctx x
  | Prop (S,x) -> Prop(S,opt {ctx with em=false; s=true} x)
  | Prop (EM,x) when ctx.s -> opt ctx x
  | Prop (EM,x) when ctx.em -> opt {ctx with em=false} x
  | Prop (EM,x) when (irrelevant_em true (opt ctx x)) -> opt {ctx with em=true} x
  | Prop (EM,x) -> Prop(EM,opt {ctx with em=true} x)

  | Txt t -> Txt (squeeze_spaces t)
  | leaf -> leaf
      
  

let rec optimize doc = opt root_ctx doc
